package ru.dbotthepony.kommons.arrays

import java.nio.ByteBuffer
import java.nio.ShortBuffer

abstract class Short2DArray : Array2D() {
	abstract operator fun get(column: Int, row: Int): Short
	abstract operator fun set(column: Int, row: Int, value: Short)

	fun storeColumnRow(target: ByteBuffer) = matrixStoreColumnRow(this, target, Short2DArray::get, ByteBuffer::putShort)
	fun storeRowColumn(target: ByteBuffer) = matrixStoreRowColumn(this, target, Short2DArray::get, ByteBuffer::putShort)
	fun storeColumnRow(target: ShortBuffer) = matrixStoreColumnRow(this, target, Short2DArray::get, ShortBuffer::put)
	fun storeRowColumn(target: ShortBuffer) = matrixStoreRowColumn(this, target, Short2DArray::get, ShortBuffer::put)
	fun store(target: Short2DArray) { target.load(this) }

	fun loadColumnRow(target: ByteBuffer) = matrixLoadColumnRow(this, target, Short2DArray::set, ByteBuffer::getShort)
	fun loadRowColumn(target: ByteBuffer) = matrixLoadRowColumn(this, target, Short2DArray::set, ByteBuffer::getShort)
	fun loadColumnRow(target: ShortBuffer) = matrixLoadColumnRow(this, target, Short2DArray::set, ShortBuffer::get)
	fun loadRowColumn(target: ShortBuffer) = matrixLoadRowColumn(this, target, Short2DArray::set, ShortBuffer::get)
	fun load(from: Short2DArray) { matrixLoad(from, this, Short2DArray::get, Short2DArray::set) }
	fun storeTransposed(from: Short2DArray) { matrixLoadTransposed(from, this, Short2DArray::get, Short2DArray::set) }

	open fun fill(value: Short) {
		for (x in 0 until rows) {
			for (y in 0 until columns) {
				this[x, y] = value
			}
		}
	}

	open fun contentEquals(other: Short2DArray): Boolean = matrixContentEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }

	override fun equals(other: Any?): Boolean {
		return matrixEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }
	}

	override fun hashCode(): Int {
		return matrixHashCode(this) { a, b, c -> a[b, c].hashCode() }
	}

	override fun toString(): String {
		return matrixToString(this, Short2DArray::get)
	}

	open val transposed: Short2DArray get() {
		return allocate(rows, columns).also { it.load(this) }
	}

	private class Impl(
		override val columns: Int,
		override val rows: Int
	) : Short2DArray() {
		private val mem = ShortArray(columns * rows)

		override fun get(column: Int, row: Int): Short {
			if (column !in 0 until columns || row !in 0 until rows)
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")

			return mem[column + row * columns]
		}

		override fun set(column: Int, row: Int, value: Short) {
			if (column !in 0 until columns || row !in 0 until rows)
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")

			mem[column + row * columns] = value
		}

		override fun fill(value: Short) {
			mem.fill(value)
		}

		override fun equals(other: Any?): Boolean {
			if (other is Impl) return sizeEquals(other) && mem.contentEquals(other.mem)
			return super.equals(other)
		}

		override fun toString(): String {
			return "Short2DArray" + super.toString()
		}
	}

	companion object {
		@JvmStatic
		fun allocate(columns: Int, rows: Int): Short2DArray {
			return Impl(columns, rows)
		}

		@JvmField
		val ZERO: Short2DArray = Impl(0, 0)
	}
}
